from datetime import timedelta
from os import PathLike
from pathlib import Path

from geojson import Point
from shapely.geometry import LineString

from .tracks import APRSTrack

KML_STANDARD = '{http://www.opengis.net/kml/2.2}'


def write_aprs_packet_tracks(packet_tracks: [APRSTrack], output_filename: PathLike):
    if not isinstance(output_filename, Path):
        output_filename = Path(output_filename)
    output_filename = output_filename.resolve().expanduser()
    extension = output_filename.suffix
    if extension == '.geojson':
        import geojson

        features = []
        for packet_track in packet_tracks:
            features.extend(
                geojson.Feature(
                    geometry=geojson.Point(packet.coordinates.tolist()),
                    properties={
                        'time': f'{packet.time:%Y%m%d%H%M%S}',
                        'callsign': packet.from_callsign,
                        'altitude': packet.coordinates[2],
                        'ascent_rate': packet_track.ascent_rates[packet_index],
                        'ground_speed': packet_track.ground_speeds[packet_index],
                    },
                )
                for packet_index, packet in enumerate(packet_track)
            )

            features.append(
                geojson.Feature(
                    geometry=geojson.LineString(
                        [packet.coordinates.tolist() for packet in packet_track.packets]
                    ),
                    properties={
                        'time': f'{packet_track.packets[-1].time:%Y%m%d%H%M%S}',
                        'callsign': packet_track.callsign,
                        'altitude': packet_track.coordinates[-1, -1],
                        'ascent_rate': packet_track.ascent_rates[-1],
                        'ground_speed': packet_track.ground_speeds[-1],
                        'seconds_to_ground': packet_track.time_to_ground
                                             / timedelta(seconds=1),
                    },
                )
            )

        features = geojson.FeatureCollection(features)

        with open(output_filename, 'w') as output_file:
            geojson.dump(features, output_file)
    elif extension == '.kml':
        from fastkml import kml

        output_kml = kml.KML()
        document = kml.Document(
            KML_STANDARD, '1', 'root document', 'root document, containing geometries'
        )
        output_kml.append(document)

        for packet_track_index, packet_track in enumerate(packet_tracks):
            for packet_index, packet in enumerate(packet_track):
                placemark = kml.Placemark(
                    KML_STANDARD,
                    f'1 {packet_track_index} {packet_index}',
                    f'{packet_track.callsign} {packet.time:%Y%m%d%H%M%S}',
                    f'altitude={packet.coordinates[2]} '
                    f'ascent_rate={packet_track.ascent_rates[packet_index]} '
                    f'ground_speed={packet_track.ground_speeds[packet_index]}',
                )
                placemark.geometry = Point(packet.coordinates.tolist())
                document.append(placemark)

            placemark = kml.Placemark(
                KML_STANDARD,
                f'1 {packet_track_index}',
                packet_track.callsign,
                f'altitude={packet_track.coordinates[-1, -1]} '
                f'ascent_rate={packet_track.ascent_rates[-1]} '
                f'ground_speed={packet_track.ground_speeds[-1]} '
                f'seconds_to_ground={packet_track.time_to_ground / timedelta(seconds=1)}',
            )
            placemark.geometry = LineString(packet_track.coordinates)
            document.append(placemark)

        with open(output_filename, 'w') as output_file:
            output_file.write(output_kml.to_string())
    else:
        raise NotImplementedError(
            f'saving to file type "{extension}" has not been implemented'
        )



from datetime import datetime, timedelta
from logging import Logger
from os import PathLike

from aprslib.packets.base import APRSPacket
import numpy

from packetraven import APRSDatabaseTable
from packetraven.base import APRSPacketSource
from packetraven.connections import TimeIntervalError
from packetraven.tracks import APRSTrack
from packetraven.utilities import get_logger
from packetraven.writer import write_aprs_packet_tracks

LOGGER = get_logger('packetraven')


def retrieve_packets(
        connections: [APRSPacketSource],
        packet_tracks: [APRSTrack],
        database: APRSDatabaseTable = None,
        output_filename: PathLike = None,
        start_date: datetime = None,
        end_date: datetime = None,
        logger: Logger = None,
) -> [APRSPacket]:
    if logger is None:
        logger = LOGGER

    logger.debug(f'receiving packets from {len(connections)} source(s)')
    current_time = datetime.now()

    parsed_packets = []
    for connection in connections:
        try:
            connection_packets = connection.packets
            parsed_packets.extend(connection_packets)
        except ConnectionError as error:
            LOGGER.error(f'{connection.__class__.__name__} - {error}')
        except TimeIntervalError:
            pass

    logger.debug(f'received {len(parsed_packets)} packets')

    if len(parsed_packets) > 0:
        updated_callsigns = set()
        new_packets = {}
        for parsed_packet in parsed_packets:
            callsign = parsed_packet['callsign']

            if start_date is not None and parsed_packet.time <= start_date:
                continue
            if end_date is not None and parsed_packet.time >= end_date:
                continue

            if callsign not in packet_tracks:
                packet_tracks[callsign] = APRSTrack(callsign, [parsed_packet])
                logger.debug(f'started tracking callsign {callsign:8}')
            else:
                packet_track = packet_tracks[callsign]
                if parsed_packet not in packet_track:
                    packet_track.append(parsed_packet)
                else:
                    if database is None or parsed_packet.source != database.location:
                        logger.debug(f'skipping duplicate packet: {parsed_packet}')
                    continue

            if parsed_packet.source not in new_packets:
                new_packets[parsed_packet.source] = []
            new_packets[parsed_packet.source].append(parsed_packet)
            updated_callsigns.add(callsign)

        parsed_packets = []
        for source, packets in new_packets.items():
            logger.info(f'received {len(packets)} new packet(s) from {source}: {packets}')
            parsed_packets.extend(packets)

        parsed_packets = sorted(parsed_packets)
        if database is not None:
            database.send(parsed_packets)

        updated_callsigns = sorted(updated_callsigns)
        for callsign in updated_callsigns:
            packet_track = packet_tracks[callsign]
            packet_time = datetime.utcfromtimestamp(
                (packet_track.times[-1] - numpy.datetime64('1970-01-01T00:00:00Z'))
                / numpy.timedelta64(1, 's')
            )

            message = ''
            try:
                coordinate_string = ', '.join(
                    f'{coordinate:.3f}°' for coordinate in packet_track.coordinates[-1, :2]
                )
                message += (
                    f'{callsign:8} #{len(packet_track)} ({coordinate_string}, {packet_track.coordinates[-1, 2]:.2f} m); '
                    f'{(current_time - packet_time) / timedelta(seconds=1)} s old; '
                    f'{packet_track.intervals[-1]:.2f} s since last packet; '
                    f'{packet_track.distances[-1]:.2f} m distance over ground ({packet_track.ground_speeds[-1]:.2f} m/s), '
                    f'{packet_track.ascents[-1]:.2f} m ascent ({packet_track.ascent_rates[-1]:.2f} m/s)'
                )

                if packet_track.time_to_ground >= timedelta(seconds=0):
                    current_time_to_ground = (
                            packet_time + packet_track.time_to_ground - current_time
                    )
                    message += (
                        f'; currently falling from max altitude of {packet_track.coordinates[:, 2].max():.3f} m; '
                        f'{current_time_to_ground / timedelta(seconds=1):.2f} s to the ground'
                    )
            except Exception as error:
                LOGGER.exception(f'{error.__class__.__name__} - {error}')
            finally:
                logger.info(message)

        if output_filename is not None:
            write_aprs_packet_tracks(
                [packet_tracks[callsign] for callsign in updated_callsigns], output_filename
            )

    return parsed_packets



from abc import ABC, abstractmethod
from datetime import timedelta

import requests
from serial.tools import list_ports

from packetraven.packets import APRSPacket, LocationPacket


class Connection(ABC):
    interval: timedelta = None

    def __init__(self, location: str):
        self.location = location

    @abstractmethod
    def close(self):
        raise NotImplementedError


class NetworkConnection(Connection):
    """
    A connection over the Internet
    """

    @property
    def connected(self) -> bool:
        """ whether current session has a network connection """
        try:
            requests.get(self.location, timeout=2)
            return True
        except (requests.ConnectionError, requests.Timeout):
            return False


class PacketSource(Connection):
    @property
    @abstractmethod
    def packets(self) -> [LocationPacket]:
        """
        most recent available location packets, since the last minimum time interval if applicable

        :return: list of location packets
        """

        raise NotImplementedError


class APRSPacketSource(PacketSource):
    def __init__(self, location: str, callsigns: [str]):
        """
        Create a new generic APRS packet connection.

        :param location: location of APRS packets
        :param callsigns: list of callsigns to return from source
        """

        super().__init__(location)
        self.callsigns = callsigns

    @property
    @abstractmethod
    def packets(self) -> [APRSPacket]:
        """ most recent available APRS packets, since the last minimum time interval (if applicable) """
        raise NotImplementedError


class PacketSink(Connection):
    @property
    @abstractmethod
    def send(self, packets: [LocationPacket]):
        """ send given packets to remote """
        raise NotImplementedError


class APRSPacketSink(PacketSink):
    @property
    def send(self, packets: [APRSPacket]):
        super().send(packets)


def next_open_serial_port() -> str:
    """
    Get next port in ports list.

    :return: port name
    """

    try:
        return next(available_serial_ports())
    except StopIteration:
        raise ConnectionError('no open serial ports')


def available_serial_ports() -> str:
    """
    Iterate over available serial ports.

    :return: port name
    """

    for com_port in list_ports.comports():
        yield com_port.device
    else:
        return None
